#include <avr/pgmspace.h>
#include <math.h>
#include "include/config.h"
#include "include/t6963c.h"
#include "include/graphics.h"

/** ******************************************
*	@brief	bit manipulation macros
**********************************************/
#define bit_get(p,m) ((p) & (m))
#define bit_set(p,m) ((p) |= (m))
#define bit_clear(p,m) ((p) &= ~(m))
#define bit_flip(p,m) ((p) ^= (m))
#define bit_write(c,p,m) (c ? bit_set(p,m) : bit_clear(p,m))
#define BIT(x) (0x01 << (x))
#define LONGBIT(x) ((unsigned long)0x00000001 << (x))

static inline void bresenham_line (glcd_x_t x1, glcd_y_t y1, glcd_x_t x2, glcd_y_t y2, uint_fast8_t show)
{	
	int16_t dx, dy, x, y;
	int8_t sdx, sdy;
	uint16_t i, px, py;

	dx = (int16_t)x2-x1;      // the horizontal distance of the line
	dy = (int16_t)y2-y1;      // the vertical distance of the line
	if (dx == 0)
		sdx = 0;
	if (dx > 0)
		sdx = 1;
	else
	{  	sdx = -1;
		dx = -dx;
	}
	if (dy == 0)
		sdy = 0;
	if (dy > 0)
		sdy = 1;
	else
	{	sdy = -1;
		dy = -dy;
	}
	x = dx * 2;
	y = dy * 2;
	px = x1;
	py = y1;
	GLCD_PIXEL(px, py, show);

	if (dx >= dy)					 // the line is more horizontal than vertical
	{	
		for(i = 0; i < dx; i++)
		{	
			y += dy;
			if (y >= dx)
			{	
				y -= dx;
				py += sdy;
			}
			px += sdx;
			GLCD_PIXEL(px, py, show);
		}
	}
	else				 		// the line is more vertical than horizontal
	{	
		for(i = 0; i < dy; i++)
		{	
			x += dx;
			if (x >= dy)
			{	
				x -= dy;
				px += sdx;
			}
			py += sdy;
			GLCD_PIXEL(px, py, show);
		}
	}
}

/** *************************************************************
	@brief	Draws a line on the Display

	@param 	x1 	: start column
	@param 	y1 	: start row
	@param 	x2 	: end column
	@param 	y2 	: end row
	@param 	c 	: TRUE or FALSE for set or unset

	@return	none
*****************************************************************/
void GLCD_LINE(glcd_x_t x1, glcd_y_t y1, glcd_x_t x2, glcd_y_t y2, uint_fast8_t show)
{   
	if (y1 == y2)
	{	
		if (x1 > x2)
		{	
			glcd_x_t temp = x1;
			x1 = x2;
			x2 = temp;
		}

		/* Da im menuesystem sehr oft horizontale Linien
           verwendet werden lohnt sich das */
		/* Fast horizontal line */ 
    	while((x1 <= x2) && (x1 % GLCD_FONT_WIDTH))
			GLCD_PIXEL(x1++, y1, show);
		GLCD_GOTO(x1, y1);
		while((x1 / GLCD_FONT_WIDTH) < (x2 / GLCD_FONT_WIDTH))
		{ 
			GLCD_DPUT(show ? 0xFF : 0); 
			GLCD_CPUT(0xC0);
			x1 += GLCD_FONT_WIDTH;
		}
   		while(x1 <= x2)
			GLCD_PIXEL(x1++, y1, show);
	}
	else if (x1 == x2)
	{	
		if (y1 > y2)
		{	
			glcd_y_t temp = y1;
			y1 = y2;
			y2 = temp;
		}
		while(y1 <= y2)
			GLCD_PIXEL(x1, y1++, show);
	}
	else
		bresenham_line(x1, y1, x2, y2, show);
}



/** *************************************************************
	@brief	Draws a xbm picture on the Display
	@param 	x 	: column
	@param 	y 	: row
	@param 	w 	: width
	@param 	h 	: height
	@param 	mode 	: can be TRANS, SOLID, INVERSE
			  modes can be mixed up
	@return	none
*****************************************************************/
void GLCD_XBM(const unsigned char *xbm, glcd_x_t x, glcd_y_t y, glcd_x_t w, glcd_y_t h, uint_fast8_t mode)
{
  uint8_t D;
  uint8_t mask, i;

  for(glcd_y_t zeile = 0; zeile < h; zeile++)
  {
	  for(glcd_x_t spalte = 0; spalte < (w/8); spalte++)
	  {
	    D = pgm_read_byte( xbm + (zeile*(w/8)+spalte) );
	    for(i = 0, mask = 1; i < 8; i++, mask <<= 1)
	    {
		  //D = bit_get(pgm_read_byte(xbm+(zeile*(w/8)+spalte)), mask /*BIT(i)*/);
		  if(D & mask)
		  {
		    if( ((mode & TRANS) && !(mode & INVERSE)) )
		    {
		      GLCD_PIXEL(x+i+(spalte*8), y+zeile, TRUE);
		    }
		    else if( ((mode & SOLID) && !(mode & INVERSE)) )
		    {
		      GLCD_PIXEL(x+i+(spalte*8), y+zeile, TRUE);
		    }
		    else if( ((mode & SOLID) && (mode & INVERSE)) )
		    {
		      GLCD_PIXEL(x+i+(spalte*8), y+zeile, FALSE);
		    }
		  }
		  else //if(!D)
		  {
		    if( (mode & INVERSE) )
		    {
		      GLCD_PIXEL(x+i+(spalte*8), y+zeile, TRUE);
		    }
		    else if( (mode & SOLID) && !(mode & INVERSE) )
		    {
		      GLCD_PIXEL(x+i+(spalte*8), y+zeile, FALSE);
		    }
		  }
	    }
	  }
    }
}

/** *************************************************************
	@brief	Draws a box on the Display
	@param 	x1 	: start column
	@param 	y1 	: start row
	@param 	x2 	: end column
	@param 	y2 	: end row
	@param 	show 	: TRUE or FALSE
	@return	none
*****************************************************************/
void GLCD_BOX(glcd_x_t x1, glcd_y_t y1, glcd_x_t x2, glcd_y_t y2, uint_fast8_t show)
{
  GLCD_LINE(x1,y1,x2,y1,show);  // up
  GLCD_LINE(x1,y2,x2,y2,show);  // down
  GLCD_LINE(x2,y1,x2,y2,show);  // right
  GLCD_LINE(x1,y1,x1,y2,show);  // left
}

/** *************************************************************
	@brief	Draws a filled box on the Display
	@param 	x1 	: start column
	@param 	y1 	: start row
	@param 	x2 	: end column
	@param 	y2 	: end row
	@param 	show 	: TRUE or FALSE
	@return	none
*****************************************************************/
void GLCD_BOX_FILLED (glcd_x_t x1, glcd_y_t y1, glcd_x_t x2, glcd_y_t y2, uint_fast8_t show)
{
	glcd_y_t i;

	for (i = y1; i <= y2; i++)
		GLCD_LINE(x1, i, x2, i, show);
}


/** *************************************************************
	@brief	Draws a degree line, like on a clock
	@param 	x 		: middle column
	@param 	y 		: middel row
	@param 	degree 		: How many degrees
	@param 	inner_radius 	: The inner start of the line
	@param 	outer_radius 	: The outer end of the line
	@param 	show 		: TRUE or FALSE
	@return	none
*****************************************************************/
/*
void GLCD_DEGREE_LINE(glcd_x_t x, glcd_y_t y, int degree, int inner_radius, int outer_radius, uint_fast8_t show)
{
  glcd_x_t fx, tx;
  glcd_y_t fy, ty;

  fx = x + (inner_radius * sin(degree * 3.14 / 180));    //  ???????????????
  fy = y - (inner_radius * cos(degree * 3.14 / 180));
  tx = x + (outer_radius * sin(degree * 3.14 / 180));
  ty = y - (outer_radius * cos(degree * 3.14 / 180));
  GLCD_LINE(fx,fy,tx,ty,show);
}


/** *************************************************************
	@brief	Draws a bold degree line, like on a clock
	@param 	x 		: middle column
	@param 	y 		: middel row
	@param 	degree 		: How many degrees
	@param 	inner_radius 	: The inner start of the line
	@param 	outer_radius 	: The outer end of the line
	@param 	show 		: TRUE or FALSE
	@return	none
*****************************************************************/
/*
void GLCD_DEGREE_LINE_BOLD(glcd_x_t x, glcd_y_t y, int degree, int inner_radius, int outer_radius, uint_fast8_t show)
{
  int fx,fy,tx,ty;

  fx = x + (inner_radius * sin(degree * 3.14 / 180));    //  ???????????????
  fy = y - (inner_radius * cos(degree * 3.14 / 180));
  tx = x + (outer_radius * sin(degree * 3.14 / 180));
  ty = y - (outer_radius * cos(degree * 3.14 / 180));
  GLCD_LINE(fx,fy,tx,ty,show);
  GLCD_LINE(fx+1,fy+1,tx+1,ty+1,show);
  GLCD_LINE(fx-1,fy-1,tx-1,ty-1,show);
}

/** *************************************************************
	@brief	Draws a circle
	@param 	x 	: middle column
	@param 	y 	: middel row
	@param 	radius 	: Radius of the circle
	@param 	show 	: TRUE or FALSE
	@return	none
*****************************************************************/
/*
void GLCD_CIRCLE(glcd_x_t x, glcd_x_t y, int radius, uint_fast8_t show)
{
  int xc = 0;
  int yc ;
  int p ;
  yc = radius;
  p = 3 - (radius << 1);
  while (xc <= yc)
  {
    GLCD_PIXEL(x + xc, y + yc, show);
    GLCD_PIXEL(x + xc, y - yc, show);
    GLCD_PIXEL(x - xc, y + yc, show);
    GLCD_PIXEL(x - xc, y - yc, show);
    GLCD_PIXEL(x + yc, y + xc, show);
    GLCD_PIXEL(x + yc, y - xc, show);
    GLCD_PIXEL(x - yc, y + xc, show);
    GLCD_PIXEL(x - yc, y - xc, show);
    if (p < 0)
      p += (xc++ << 2) + 6;
    else
      p += ((xc++ - yc--)<<2) + 10;
  }
}
*/
