#include <stdio.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <math.h>

#include "include/config.h"
#include "include/t6963c.h"

static inline uint8_t GLCD_SGET(void);

/** *************************************************************

	@brief	Write data to Display in continuous mode.
	@param 	byte : The byte to write
	@return	nothing
*****************************************************************/
void GLCD_C_DPUT(uint8_t byte)
{
  GLCD_SET_DATADIR_IN();

  GLCD_CD_HIGH();		// bring LCD C/D line high (read status byte)

  do {} while ((0x08 & GLCD_SGET()) != 0x08); // wait until display ready

  GLCD_SET_DATADIR_OUT(); 	// set data direction to output

  GLCD_CD_LOW();
  GLCD_DATA_PORT = byte;        // write value to data port
//  GLCD_RD_HIGH();		// bring LCD /RD line low (read active)
  GLCD_WR_LOW();         	// activate LCD's write mode
  GLCD_CE_LOW();                // pulse enable LOW > 80 ns

  asm volatile ("nop"::);
  asm volatile ("nop"::);

  GLCD_CE_HIGH();               // return enable HIGH
  GLCD_WR_HIGH();               // restore Write mode to inactive
}

/** *************************************************************
	@brief	Read data from Display in continuous mode.
	@param 	none
	@return	Byte
*****************************************************************/
uint8_t GLCD_C_DGET(void)
{
  uint8_t glcd_byte;

  GLCD_SET_DATADIR_IN();

  GLCD_CD_HIGH();		// bring LCD C/D line high (read status byte)

  do {} while ((0x04 & GLCD_SGET()) != 0x04); // wait until display ready

  GLCD_CD_LOW();   				// data mode
  GLCD_RD_LOW();   				// activate READ mode
  GLCD_CE_LOW();   				// enable chip, which outputs data

  asm volatile ("nop"::);			// wait 150ns
  asm volatile ("nop"::);
  asm volatile ("nop"::);
  asm volatile ("nop"::);

  glcd_byte = GLCD_DATA_PIN;  			// read data from LCD

  GLCD_CE_HIGH();   				// disable chip
  GLCD_RD_HIGH();   				// turn off READ mode

  GLCD_SET_DATADIR_OUT();

  return(glcd_byte);
}

/** *************************************************************
	@brief	Set the Address
	Will be only used internal
	@param 	addr : Addresse
	@return	none
*****************************************************************/
static void GLCD_SET_ADDRESS(uint16_t addr)
{
  GLCD_DPUT(addr);
  GLCD_DPUT((addr & 0xFF00) >> 8);
  GLCD_CPUT(0x24);
}


/** *************************************************************
	@brief	Clears the display ram
	@param 	none
	@return	none
*****************************************************************/
void GLCD_CLEAR(void)
{
  uint_fast16_t i;

  GLCD_SET_ADDRESS(GLCD_G_BASE);				// addrptr at address G_BASE

#if 0
  for (i = 0; i < (GLCD_BYTES_PER_ROW*GLCD_YMAX*GLCD_MAX_PAGES); i++)	// clear all pages
  {
    GLCD_DPUT(0);
    GLCD_CPUT(0xc0);
  }
#else
  GLCD_CPUT(0xb0);
  for (i = 0; i < (GLCD_BYTES_PER_ROW*GLCD_YMAX*GLCD_MAX_PAGES); i++)	// clear all pages
  {
    GLCD_C_DPUT(0);
  }
  GLCD_CPUT(0xb2);
#endif
}


/** *************************************************************
	@brief	Set one pixel on display
	@param 	x    : Column
	@param 	y    : Row
	@param 	show : 0=Unset, 1=set
	@return	none
*****************************************************************/
void GLCD_PIXEL(glcd_x_t x, glcd_y_t y, uint8_t show)
{
  uint16_t addr;
     									// memory address of byte containing pixel to write
  addr =  GLCD_G_BASE + (y*GLCD_BYTES_PER_ROW)  + (x/GLCD_FONT_WIDTH);
  GLCD_SET_ADDRESS(addr);   							// set LCD addr. pointer
  if(show) show = 0xf8; else show = 0xf0;
  GLCD_CPUT(show | ((GLCD_FONT_WIDTH - 1 - x % GLCD_FONT_WIDTH)) );  			// set bit-within-byte command
}


/** *************************************************************
	@brief	Setup the display
	@param 	none
	@return	none
*****************************************************************/
void GLCD_INIT(void)
{
  GLCD_SET_DATADIR_OUT();

  DDR(GLCD_CE_PORT)  |= _BV(GLCD_CE_PIN);
  DDR(GLCD_RD_PORT)  |= _BV(GLCD_RD_PIN);
  DDR(GLCD_WR_PORT)  |= _BV(GLCD_WR_PIN);
  DDR(GLCD_CD_PORT)  |= _BV(GLCD_CD_PIN);
  DDR(GLCD_RST_PORT) |= _BV(GLCD_RST_PIN);
  DDR(GLCD_FS_PORT)  |= _BV(GLCD_FS_PIN);

  GLCD_CE_HIGH();
  GLCD_RD_HIGH();
  GLCD_WR_HIGH();
  GLCD_CD_HIGH();
  GLCD_RST_HIGH();
  GLCD_FS_HIGH();


  GLCD_RST_LOW();
  _delay_ms(10);
  GLCD_RST_HIGH();

  GLCD_FS_LOW();
  _delay_ms(10);

  GLCD_DPUT(GLCD_G_BASE);
  GLCD_DPUT(GLCD_G_BASE >> 8);
  GLCD_CPUT(0x42);       // set graphics memory to address glcd_G_BASE

  GLCD_DPUT(GLCD_BYTES_PER_ROW);
  GLCD_DPUT(GLCD_BYTES_PER_ROW >> 8);
  GLCD_CPUT(0x43);  // n bytes per graphics line

  GLCD_CPUT(0x98);  // Graphics only
  GLCD_CPUT(0x80);  // Set mode
}

/** *************************************************************
	@brief	Get status byte of Display
	@param 	none
	@return	status
*****************************************************************/
static inline uint8_t GLCD_SGET(void)
{
  uint8_t glcd_status;

  GLCD_RD_LOW();		// bring LCD /RD line low (read active)
  GLCD_CE_LOW();		// bring LCD /CE line low (chip-enable active)

  asm volatile ("nop"::);
  asm volatile ("nop"::);

  glcd_status = GLCD_DATA_PIN;  // read LCD status byte

  GLCD_CE_HIGH(); 	        // bring LCD /CE line high, disabling it
  GLCD_RD_HIGH();       	// deactivate LCD read mode

  return(glcd_status);
}

static inline void GLCD_WAIT_READY(void)
{
  GLCD_SET_DATADIR_IN();

  GLCD_CD_HIGH();		// bring LCD C/D line high (read status byte)

  do {} while ((0x03 & GLCD_SGET()) != 0x03); // wait until display ready

  GLCD_SET_DATADIR_OUT(); 	// set data direction to output
}


/** *************************************************************
	@brief	Write a data byte to Display
	@param 	byte : The byte to write
	@return	status
*****************************************************************/
void GLCD_DPUT(uint8_t byte)
{
  GLCD_WAIT_READY();

  GLCD_CD_LOW();
  GLCD_DATA_PORT = byte;        // write value to data port
//  GLCD_RD_HIGH();		// bring LCD /RD line low (read active)
  GLCD_WR_LOW();         	// activate LCD's write mode
  GLCD_CE_LOW();                // pulse enable LOW > 80 ns

  asm volatile ("nop"::);
  asm volatile ("nop"::);

  GLCD_CE_HIGH();               // return enable HIGH
  GLCD_WR_HIGH();               // restore Write mode to inactive

}


/** *************************************************************
	@brief	Read a data byte from Display
	@param 	none
	@return	Byte
*****************************************************************/
uint8_t GLCD_DGET(void)
{
  uint8_t glcd_byte;

  GLCD_WAIT_READY();

  GLCD_SET_DATADIR_IN();			// port as input port
//  GLCD_WR_HIGH();   				// make sure WRITE mode is inactive
  GLCD_CD_LOW();   				// data mode
  GLCD_RD_LOW();   				// activate READ mode
  GLCD_CE_LOW();   				// enable chip, which outputs data

  asm volatile ("nop"::);			// wait 150ns
  asm volatile ("nop"::);
  asm volatile ("nop"::);
  asm volatile ("nop"::);

  glcd_byte = GLCD_DATA_PIN;  			// read data from LCD

  GLCD_CE_HIGH();   				// disable chip
  GLCD_RD_HIGH();   				// turn off READ mode

  GLCD_SET_DATADIR_OUT();

  return(glcd_byte);
}

uint8_t GLCD_GETBYTE_noinc(void)
{
	GLCD_CPUT(0xC5);								// 0xC5 = Lese Display ohne adresse zu aendern
	return GLCD_DGET();
}

void GLCD_PUTBYTE_inc(uint8_t byte)
{
	GLCD_DPUT(byte);					// Den buffer ausgeben
	GLCD_CPUT(0xC0);					// 0xC0 = Schreibe und erhoehe die adresse
}

/** *************************************************************
	@brief	Write a command byte to Display
	@param 	byte : The byte to write
	@return	status
*****************************************************************/
void GLCD_CPUT(uint8_t byte)
{
  GLCD_WAIT_READY();

  GLCD_DATA_PORT = byte;  			// present data to LCD on PC's port pins

  GLCD_CD_HIGH();         			// control/status mode
//  GLCD_RD_HIGH();         			// make sure LCD read mode is off
  GLCD_WR_LOW();         			// activate LCD write mode
  GLCD_CE_LOW();         			// pulse ChipEnable LOW, > 80 ns, enables LCD I/O

  asm volatile ("nop"::);
  asm volatile ("nop"::);

  GLCD_CE_HIGH();         			// disable LCD I/O
  GLCD_WR_HIGH();         			// deactivate write mode
}


/** *************************************************************
	@brief	Set graphic cursor
	@param 	x : The column
	@param 	y : The row
	@return	status
*****************************************************************/
void GLCD_GOTO(glcd_x_t x, glcd_y_t y)
{
  uint16_t addr;       							// memory address

  addr =  GLCD_G_BASE + (y*GLCD_BYTES_PER_ROW)  + (x/GLCD_FONT_WIDTH);
  GLCD_SET_ADDRESS(addr);   						// set LCD addr. pointer
}

#if 0
/** *************************************************************
	@brief	Set a page active
	@param 	page : The page
	@return	status
*****************************************************************/
void GLCD_SET_PAGE(uint_fast8_t page)
{
  uint16_t x = GLCD_G_BASE + (GLCD_BYTES_PER_ROW * GLCD_YMAX * page);

  GLCD_DPUT(x);
  GLCD_DPUT((x & 0xFF00) >> 8);
  GLCD_CPUT(0x42);
}
#endif

/** *************************************************************
	@brief	Clear one page in display ram
	@param 	page: The page to clear
	@return	none
*****************************************************************/
void GLCD_CLEAR_PAGE(uint_fast8_t page)
{
  uint16_t i;

  GLCD_SET_ADDRESS(GLCD_G_BASE);    								 // addrptr at address G_BASE
#if 0
  for (i = (GLCD_BYTES_PER_ROW*GLCD_YMAX*page); i < (GLCD_BYTES_PER_ROW*GLCD_YMAX*(page+1))-1; i++)
  {
    GLCD_DPUT(0);
    GLCD_CPUT(0xc0);
  }
#else
  GLCD_CPUT(0xb0);
  for (i = (GLCD_BYTES_PER_ROW*GLCD_YMAX*page); i < (GLCD_BYTES_PER_ROW*GLCD_YMAX*(page+1))-1; i++)
  {
    GLCD_C_DPUT(0);
  }
  GLCD_CPUT(0xb2);
#endif
}

#if 0
/** *************************************************************
	@brief	Copy a page to another one
	@param 	source	: Page to copy
	@param 	dest	: Destination page
	@return	none
*****************************************************************/
void GLCD_COPY_PAGE(uint_fast8_t source, uint_fast8_t dest)
{
#if 1
  glcd_x_t x;
  glcd_y_t y;
  char buffer;

  for(y = 0; y < GLCD_YMAX; y++)
  {
    for(x = 0; x < GLCD_XMAX; x += GLCD_FONT_WIDTH)
    {
      GLCD_GOTO(x, y+(GLCD_YMAX*source));
      GLCD_CPUT(0xC5);					// 0xC5 = Read one byte
      buffer = GLCD_DGET();
      GLCD_GOTO(x, y+(GLCD_YMAX*dest));
      GLCD_DPUT(buffer);
      GLCD_CPUT(0xC4);					// 0xC4 = write one byte
    }
  }
#else
  int x;
  glcd_y_t y;
  char buffer[GLCD_BYTES_PER_ROW];

  for(y = 0; y < GLCD_YMAX; y++)
  {
      /* Read line of pixel */
      GLCD_GOTO(0, y+(GLCD_YMAX*source));
      GLCD_CPUT(0xb1);
      for(x = 0; x < GLCD_BYTES_PER_ROW; x++)
      {
	buffer[x] = GLCD_C_DGET();
      }
      GLCD_CPUT(0xb2);

      /* Write line of pixel */
      GLCD_GOTO(0, y+(GLCD_YMAX*dest));
      GLCD_CPUT(0xb0);
      for(x = 0; x < GLCD_BYTES_PER_ROW; x++)
      {
    	GLCD_C_DPUT(buffer[x]);
      }
      GLCD_CPUT(0xb2);
  }
#endif
}
#endif
